
/*
 * MGL graphics library
 * Copyright (C) 1999-2000, ing. Marek Gayer

 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.

 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <mgl.h>

#ifndef GL_BGR
#ifdef GL_BGR_EXT
  #define GL_BGR GL_BGR_EXT
#else
  #define GL_BGR 0x80E0
#endif
#endif

#ifndef GL_BGRA
#ifdef GL_BGRA_EXT
  #define GL_BGRA GL_BGRA_EXT
#else
  #define GL_BGRA 0x80E1
#endif
#endif

extern int mglExtBGRA;
extern int mglExtTexture;
extern int mglVersionNumber;

static int alphaExtend;

#define MGL_IMAGEFORMAT_ERR 0
#define MGL_IMAGEFORMAT_BMP 1
#define MGL_IMAGEFORMAT_SGI 2
#define MGL_IMAGEFORMAT_TGA 3

#define MGL_IMAGETYPE_NONE 0
#define MGL_IMAGETYPE_L 1
#define MGL_IMAGETYPE_LA 2
#define MGL_IMAGETYPE_RGB 3
#define MGL_IMAGETYPE_RGBA 4

typedef struct {
  unsigned char r;
  unsigned char g;
  unsigned char b;
  unsigned char a;
} MGL_RGBAColor;

typedef struct {
  unsigned char r;
  unsigned char g;
  unsigned char b;
} MGL_RGBColor;

typedef struct {
  unsigned char l;
  unsigned char a;
} MGL_LAColor;


// these variables enables to test the byte encoding on different platforms
// intel pc - bytes are swapped
// motorola appple and other platforms - bytes are stored in logical form
static short testSwapBytes = 1;
static char *pcByte = (char *) &testSwapBytes;
//static char *appleByte = (((char *) &testSwapBytes) + 1);

#define swapShortBytes(x) ((((x) & 0xff) << 8) | (((x) & 0xff00) >> 8))
#define swapLongBytes(x) (((((x) & 0xff) << 24) | (((x) & 0xff00) << 8)) | ((((x) & 0xff0000) >> 8) | (((x) & 0xff000000) >> 24)))

unsigned short pcShort (unsigned short x)
{
  if (*pcByte)
    return x;
  else
    return swapShortBytes (x);
}

unsigned long pcLong (unsigned long x)
{
  if (*pcByte)
    return x;
  else
    return swapLongBytes (x);
}


/*---------------------------------------------------------------------------*/
/* Private helper data and functions */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/

char *curFileName;

void imageError (int error, char *s)
{
  char err [1000];

  if (curFileName)
  {
    sprintf (err, "%s (%s)", s, curFileName);
    mglError (error, err);
  }
  else
  {
    mglError (error, s);
  }
}


typedef struct
{
  int pos;
  char *id;
  int len;
  int type;
} Tdet;

Tdet detect [] = {
{0, "BM", 2, MGL_IMAGEFORMAT_BMP},
{0, "\x01\xDA", 2, MGL_IMAGEFORMAT_SGI},
{0, "\x0\x0\x3\x0\x0\x0\x0", 7, MGL_IMAGEFORMAT_TGA},
{0, "\x0\x0\x2\x0\x0\x0\x0", 7, MGL_IMAGEFORMAT_TGA},
{0, "\x0\x1\x1\x0\x0\x0", 6, MGL_IMAGEFORMAT_TGA},
};

#define detectBufferLen (sizeof (detect) / sizeof (detect [0]))

static int detectFormat (FILE* f)
{
	char buf [32];
  int i;

  long pos;

  pos = ftell (f);

	if ((pos == -1L) || (fread (&buf, sizeof (buf), 1, f ) != 1) ||
    (fseek( f, pos, SEEK_SET ) != 0 ) )
  {
		imageError (MGL_ERROR_READING_TGA_HEADER, "Error reading TGA image header");
		return 0;
	}

	for (i = 0; i < detectBufferLen; i++)
  {
	  if (memcmp (buf + detect [i].pos, detect [i].id, detect [i].len) == 0)
    {
		  return detect [i].type;
    }
	}
  return 0;
}



mglImage *mglNewImage (void)
{
  mglImage *image;

  image = _malloc (sizeof (mglImage));
  image->w = image->h = image->d = 0;
  image->xrev = image->yrev = 0;
  image->type = 0;
  image->data = NULL;
  image->mipmap = NULL;
  return image;
}

void mglDeleteImage (mglImage *image)
{
  if (image != NULL)
  {
    if (image->data != NULL)
      _free (image->data);
    image->w = image->h = image->d = 0;
    image->xrev = image->yrev = 0;
    image->type = 0;
    image->data = NULL;
    _free (image);
  }
}


mglImage *mglCreateImage (int width, int height, int depth)
{
  mglImage *image;

  mglNote ("Allocating image\n");
  image = mglNewImage ();

  image->w = width;
  image->h = height;
  image->d = depth;
  //image->c = depth;

  mglNote ("Allocating image data\n");
  image->data = (unsigned char *) _malloc (width * height * depth + 4);
  mglNote ("Succesfull\n");

  return image;
}


/*---------------------------------------------------------------------------*/
/* Truevision TGA */
/*---------------------------------------------------------------------------*/

typedef struct {
	unsigned char id_length; /* Delka Image Identification Field */
	unsigned char palette_type; /* == 1 pro colormapped  0 no palette */
	unsigned char image_type; /* == 2 pro RGB, 3 pro grayscale 1 palette */
  unsigned short first_color; /* unaligned short */
	unsigned short palette_colors;
  unsigned char palette_entry_size;
	unsigned short left;
	unsigned short top;
	unsigned short image_width;
  unsigned short image_height;
	unsigned char pixel_bits;
	unsigned char descriptor_bits;
} TGA_header;

typedef struct {
  long unknown1;
  long unknown2;
} TGA_ending;

#define TGA_NORMALIMAGE 0x20
#define TGA_headerSize 18

/*---------------------------------------------------------------------------*/


typedef void (*convertFunc) (char *to, char *from, int count);

static void rgb2rgb (char *to, char *from, int count)
{
  memcpy (to, from, 3*count);
  return;
}

static void rgba2rgba (char *to, char *from, int count)
{
  memcpy (to, from, 4*count);
  return;
}

#define readconvert_bgrx2bgrx readconvert_rgbx2rgbx
static void readconvert_rgbx2rgbx (mglImage *image, FILE *f)
{
  // direct BGR to BGR and BGRA to BGRA copy

  mglNote ("Reading image with no conversion needed\n");
  if (fread (image->data, image->w * image->h, image->d, f) != (unsigned) image->d)
  {
    imageError (MGL_ERROR_READING_IMAGE_DATA, "Error reading image data");
  }
}

char *mglBufAlloc (int *bufferSize, int size, int numItems)
{
  char *p;
  int bytes;
  int maxBufferSize =  1024 * 1024;

  bytes = size * numItems;
  if (bytes > maxBufferSize)
  {
    bytes = maxBufferSize;
  }

  do
  {
    p = malloc (bytes);
    *bufferSize = bytes;
    bytes /= 2;
    if (bytes == 0)
      imageError (MGL_ERROR_NOT_ENOUGH_MEMORY, "Not enough memory for allocation of temporary file buffer");
  }
  while (p == NULL);

  *bufferSize /= size;

  return p;
}



static void bgra2rgba (char *to, char *from, int count)
{
  // converting BGRA to RGBA
  int i;

  for (i = 0; i < count; i++)
  {
//    *(long *) to = *(long *) from;
    to [0] = from [2];
    to [1] = from [1];
    to [2] = from [0];
    to [3] = from [3];
    to += 4;
    from += 4;
  }
}

static void rgba2bgra (char *to, char *from, int count)
{
	bgra2rgba (to, from, count);
}

static void bgr2rgb (char *to, char *from, int count)
{
  int i;

  for (i = 0; i < count; i++)
  {
    to [0] = from [2];
    to [1] = from [1];
    to [2] = from [0];
    to += 3;
    from += 3;
  }
}

static void rgb2bgr (char *to, char *from, int count)
{
	bgr2rgb (to, from, count);
}


static void bgr2bgra (char *to, char *from, int count)
{
  int i;

  for (i = 0; i < count; i++)
  {
    *(int *) to = *(int *) from;
    to [3] = 0;
    to += 4;
    from += 3;
  }
}



static void bgra2bgr (char *to, char *from, int count)
{
  // converting BGRA to BGR
  int i;

  for (i = 0; i < count ; i++)
  {
    *(long *) to = *(long *) from;
    to += 3;
    from += 4;
  }
}

static void bgra2rgb (char *to, char *from, int count)
{
  int i;

  for (i = 0; i < count ; i++)
  {
    to [0] = from [2];
    to [1] = from [1];
    to [2] = from [0];
    *to += 3;
    from += 4;
  }
}


static void bgr2rgba (char *to, char *from, int count)
{
  int i;

  for (i = 0; i < count; i++)
  {
    to [0] = from [2];
    to [1] = from [1];
    to [2] = from [0];
    to [3] = 0;
    to += 4;
    from += 3;
  }
}




static void l2la (char *to, char *from, int count)
{
  int i;

  for (i = 0; i < count; i++)
  {
    *(char *) to = *(char *) from;
    to++;
    *to = 0;
    to++;
    from++;
  }
}

static void la2l (char *to, char *from, int count)
{
  int i;

  for (i = 0; i < count; i++)
  {
    *(char *) to = *(char *) from;
    to++;
    from++;
    from++;
  }
}


static void readconvert (mglImage *image, FILE *f, convertFunc func)
{
  // converting BGRA to BGR
  int wh;
  unsigned char *data;
  int bufferSize;
  int count;
  char *buf;
  int to, from;

  to = from = 0;

  if ((func == bgr2bgra) || (func == bgr2rgba))
    {from = 3; to = 4;}
  if (func == bgr2rgb)
    {from = 3; to = 3;}

  if (func == l2la)
    {from = 1; to = 2;}

  if ((func == bgra2rgb) || (func == bgra2bgr))
    {from = 4; to = 3;}
  if (func == bgra2rgba)
    {from = 4; to = 4;}

  wh = image->w * image->h;
  data = image->data;
  buf = mglBufAlloc (&bufferSize, from, wh);

  while (wh)
  {
    count = fread (buf, from, bufferSize, f);
    if (count == 0)
    {
      imageError (MGL_ERROR_READING_IMAGE_DATA, "Error reading image data");
    }
    wh -= count;

    func ((char *)data, buf, count);
    data += count * to;
  }

  free (buf);
}


static void mglDetectGrayscaleTGA (TGA_header *h, int alpha, int *tgaDepth, int *imageDepth)
{
  if (h->pixel_bits != 8 )
  {
		imageError (MGL_ERROR_UNSUPPORTED_TGA_IMAGE, "Recognized uncompressed grayscale TGA image must have 8 bits per pixel");
  }
	else
  {
    *tgaDepth = 1;

    switch (alpha)
    {
      case 0:
        *imageDepth = 1;
        break;
      case 1:
        *imageDepth = 2;
        break;
      default:
        if (mglVersionNumber < 11)
        {
          *imageDepth = 2;
          alphaExtend = 1;
        }
        else
          *imageDepth = 1;
    }
	}
}

static void mglDetectPaletteTGA (TGA_header *h, int alpha, int *tgaDepth, int *imageDepth)
{
  if (h->pixel_bits != 8 )
  {
		imageError (MGL_ERROR_UNSUPPORTED_TGA_IMAGE, "Recognized uncompressed palette TGA image must have 8 bits per pixel");
  }
	else
  {
    *tgaDepth = 1;

    switch (alpha)
    {
      case 0:
        *imageDepth = 1;
        break;
      case 1:
        *imageDepth = 2;
        break;
      default:
        if (mglVersionNumber < 11)
        {
          *imageDepth = 2;
          alphaExtend = 1;
        }
        else
          *imageDepth = 1;
    }
	}
}

static void mglDetectRGBTGA (TGA_header *h, int alpha, int *tgaDepth, int *imageDepth)
{
  if (h->pixel_bits == 24)
  {
		*tgaDepth = 3;

    if (alpha >= 2)
    {
      alphaExtend = 1;
      alpha = 1;
    }

    *imageDepth = alpha + 3;
  }
	else
  {
		if (h->pixel_bits == 32 )
    {
			if (h->descriptor_bits != 8 )
      {
		    imageError (MGL_ERROR_UNSUPPORTED_TGA_IMAGE, "Recognized uncompressed 32-bit RGBA TGA image must have 8 descriptor bits");
			}
			*tgaDepth = *imageDepth = 4;
		}
		else
    {
		  imageError (MGL_ERROR_UNSUPPORTED_TGA_IMAGE, "Recognized uncompressed RGB TGA image must have 24 or 32 bits per pixel");
    }
	}
}

static void mglDetectTGAType (TGA_header *h, int alpha, int *tgaDepth, int *imageDepth)
{
  // checkup tga header type
	switch (h->image_type)
  {
    case 2:
      // detected tga is RGB
      mglDetectRGBTGA (h, alpha, tgaDepth, imageDepth);
      break;
	case 1:
		mglDetectPaletteTGA (h, alpha, tgaDepth, imageDepth);
		break;
    case 3:
      // detected tga is grayscale
      mglDetectGrayscaleTGA (h, alpha, tgaDepth, imageDepth);
      break;
    default:
	    /* TODO: Existuje i 16 bit TGA */
	    mglError (MGL_ERROR_UNSUPPORTED_TGA_IMAGE, "File is not a valid uncompressed 24 or 32 bit RGB or 8-bit grayscale TGA image");
      break;
  }
}



static void mglReadTGAHeader (TGA_header *h, FILE *f)
{
  char buf [TGA_headerSize];
  char *b = buf;

  if (fread (buf, TGA_headerSize, 1, f) != 1)
  {
		imageError (MGL_ERROR_READING_TGA_HEADER, "Error reading TGA header");
	}
  else
  {
    h->id_length = *b++;
    h->palette_type = *b; b++;
    h->image_type = *b; b++;
    h->first_color = pcShort (*(unsigned short *)b); b+=2;
    h->palette_colors = pcShort (*(unsigned short *)b); b+=2;
    h->palette_entry_size = *b; b++;
    h->left = pcShort (*(unsigned short *)b); b+=2;
    h->top = pcShort (*(unsigned short *)b); b+=2;
    h->image_width = pcShort (*(unsigned short *)b); b+=2;
    h->image_height = pcShort (*(unsigned short *)b); b+=2;
    h->pixel_bits = *b; b++;
    h->descriptor_bits = *b; b++;
  }
}

static void mglReadBGR (mglImage *image, FILE *f, int fileDepth)
{
  // checkup wheter the whole image could be read directly into image memory

  switch (image->d)
  {
    case 1:
      mglNote ("Type luminance\n");
      image->type = GL_LUMINANCE;
      readconvert_bgrx2bgrx (image, f);
      return;
    case 2:
      image->type = GL_LUMINANCE_ALPHA;
      readconvert (image, f, l2la);
      return;
  }

//  #ifdef GL_EXT_bgra
  if (mglExtBGRA)
  {
    if (fileDepth == image->d)
    {
      switch (fileDepth)
      {
      case 4:
        image->type = GL_BGRA;
        break;
      case 3:
        image->type = GL_BGR;
        break;
      }
      readconvert_bgrx2bgrx (image, f);
      return;
    }
    else
    {
      if (fileDepth == 3)
      {
        image->type = GL_BGRA;
        readconvert (image, f, bgr2bgra);
        return;
      }
      else
      {
        image->type = GL_BGR;
        readconvert (image, f, bgra2bgr);
        return;
      }
    }
  }
//  #endif

  if (fileDepth == image->d)
  {
    if (fileDepth == 4)
    {
      image->type = GL_RGBA;
      readconvert (image, f, bgra2rgba);
      return;
    }
    else
    {
      image->type = GL_RGB;
      readconvert (image, f, bgr2rgb);
      return;
    }
  }
  else
  {
    if (fileDepth == 3)
    {
      image->type = GL_RGBA;
      readconvert (image, f, bgr2rgba);
      return;
    }
    else
    {
      image->type = GL_RGB;
      readconvert (image, f, bgra2rgb);
      return;
    }
  }
}


static mglImage *loadTGA (FILE* f, int alpha)
{
	TGA_header h;
	int tgaDepth;
  int imageDepth;
  mglImage *image;

  mglNote ("Reading TGA header\n");
  mglReadTGAHeader (&h, f);

  mglNote ("Detecting TGA type\n");
  mglDetectTGAType (&h, alpha, &tgaDepth, &imageDepth);

  mglNote ("Creating memory for TGA image\n");
  image = mglCreateImage (h.image_width, h.image_height, imageDepth);

  if (h.descriptor_bits & TGA_NORMALIMAGE)
    image->xrev = image->yrev = 0;
  else
  {
    image->xrev = 0; image->yrev = 1;
  }

  mglNote ("Reading TGA image\n");
  mglReadBGR (image, f, tgaDepth);

  return image;
}

typedef struct
{
  unsigned short type;
	unsigned long size;
	unsigned short reserved1;
  unsigned short reserved2;
	unsigned long offBits;
} MGL_BITMAPFILEHEADER;

#define SIZEOF_MGL_BITMAPFILEHEADER 14


#define BI_RGB 0

typedef struct
{
	unsigned long size;
	unsigned long width;
	unsigned long height;
	unsigned short planes;
  unsigned short bitCount;
	unsigned long compression;
	unsigned long sizeImage;
	unsigned long xPelsPerMeter;
	unsigned long yPelsPerMeter;
	unsigned long clrUsed;
  unsigned long clrImportant;
} MGL_BITMAPINFOHEADER;

typedef struct
{
  MGL_BITMAPFILEHEADER bf;
  MGL_BITMAPINFOHEADER bi;
} BMP_header;

#define BMP_headerSize 54

static void mglReadBMPHeader (BMP_header *h, FILE *f)
{
  char buf [BMP_headerSize];
  char *b = buf;


  if (fread (buf, BMP_headerSize, 1, f) != 1)
  {
    mglError (MGL_ERROR_READING_BMP_HEADER, "Error reading BMP image header");
	}
  else
  {
	  h->bf.type = pcShort (*(unsigned short *)b); b+=2;
	  h->bf.size = pcLong (*(unsigned long *)b); b+=4;
	  h->bf.reserved1 = pcShort (*(unsigned short *)b); b+=2;
	  h->bf.reserved2 = pcShort (*(unsigned short *)b); b+=2;
	  h->bf.offBits = pcLong (*(unsigned long *)b); b+=4;
	  h->bi.size = pcLong (*(unsigned long *)b); b+=4;
	  h->bi.width = pcLong (*(unsigned long *)b); b+=4;
	  h->bi.height = pcLong (*(unsigned long *)b); b+=4;
	  h->bi.planes = pcShort (*(unsigned short *)b); b+=2;
	  h->bi.bitCount = pcShort (*(unsigned short *)b); b+=2;
	  h->bi.compression = pcLong (*(unsigned long *)b); b+=4;
	  h->bi.sizeImage = pcLong (*(unsigned long *)b); b+=4;
	  h->bi.xPelsPerMeter = pcLong (*(unsigned long *)b); b+=4;
	  h->bi.yPelsPerMeter = pcLong (*(unsigned long *)b); b+=4;
	  h->bi.clrUsed = pcLong (*(unsigned long *)b); b+=4;
	  h->bi.clrImportant = pcLong (*(unsigned long *)b); b+=4;
  }

  if (fseek (f, SIZEOF_MGL_BITMAPFILEHEADER + h->bi.size, SEEK_SET) != 0)
  {
		imageError (MGL_ERROR_UNSUPPORTED_BMP_IMAGE, "Error seeking to start of BMP image data");
	}
}

static int checkupGrayscalePal (FILE *f)
{
  MGL_RGBAColor pal [256];
  int i;

  if (fread ((void *)pal, 4, 256, f) != 256)
  {
    imageError (MGL_ERROR_READING_IMAGE_DATA, "Error reading 8-bit BMP pallete information");
  }

  for (i = 0; i < 256; i++)
  {
    if ((pal [i].r != i) || (pal [i].g != i) || (pal [i].b != i))
      return 0;
  }
  return 1;
}

static int mglDetectBMPType (BMP_header *h, FILE *f)
{
  int depth;

  if (h->bi.compression != 0)
  {
		imageError (MGL_ERROR_UNSUPPORTED_BMP_IMAGE, "Compressed BMP not supported");
		return 0;
	}

	if (h->bi.planes != 1)
  {
		imageError (MGL_ERROR_UNSUPPORTED_BMP_IMAGE, "BMP with more than 1 plane not supported");
		return 0;
	}

  switch (h->bi.bitCount)
  {
    case 8:
      if (((h->bi.clrUsed & 0xFF) != 0) || (!checkupGrayscalePal (f)))
      {
        imageError (MGL_ERROR_UNSUPPORTED_BMP_IMAGE, "No support for 256 colors BMP images");
      }
      depth = 1;
      break;
    case 24:
      depth = 3;
      break;
    default:
      imageError (MGL_ERROR_UNSUPPORTED_BMP_IMAGE, "No support for BMP images other than 24-bit RGB and 8-bit grayscale");
      depth = 0;
	}
  return depth;
}


mglImage *loadBMP (FILE* f, int alpha)
{
	BMP_header h;
  mglImage *image;
  int depth;

	mglReadBMPHeader (&h, f);

  depth = mglDetectBMPType (&h, f);

  if (depth == 1)
  {
    if (alpha >= 2)
    {
      if (mglVersionNumber >= 11)
        alpha = 0;
      else
      {
        alpha = 1;
        alphaExtend = 1;
      }
    }
  }
  else
  {
    if (alpha >= 2)
    {
      alphaExtend = 1;
      alpha = 1;
    }
  }

	image = mglCreateImage (h.bi.width, h.bi.height, depth + alpha);

  // BMP images are stored with y-reversed form
  image->xrev = 0;
  image->yrev = 1;

  mglReadBGR (image, f, depth);

  return image;
}


/* routines for reading .sgi and .bw */

static void
bwtorgba(unsigned char *b,unsigned char *l,int n) {
    while(n--) {
	l[0] = *b;
	l[1] = *b;
	l[2] = *b;
	l[3] = 0xff;
	l += 4; b++;
    }
}

static void
latorgba(unsigned char *b, unsigned char *a,unsigned char *l,int n) {
    while(n--) {
	l[0] = *b;
	l[1] = *b;
	l[2] = *b;
	l[3] = *a;
	l += 4; b++; a++;
    }
}

static void
rgbtorgba(unsigned char *r,unsigned char *g,unsigned char *b,unsigned char *l,int n) {
    while(n--) {
	l[0] = r[0];
	l[1] = g[0];
	l[2] = b[0];
	l[3] = 0xff;
	l += 4; r++; g++; b++;
    }
}

static void
rgbatorgba(unsigned char *r,unsigned char *g,unsigned char *b,unsigned char *a,unsigned char *l,int n) {
    while(n--) {
	l[0] = r[0];
	l[1] = g[0];
	l[2] = b[0];
	l[3] = a[0];
        l += 4; r++; g++; b++; a++;
    }
}

static void
bwtobw(unsigned char *b,unsigned char *l,int n) {
    while(n--) {
	l[0] = *b;
	l += 1; b++;
    }
}

static void
latola(unsigned char *b, unsigned char *a,unsigned char *l,int n) {
    while(n--) {
	l[0] = *b;
	l[1] = *a;
	l += 2; b++; a++;
    }
}

static void
rgbtorgb(unsigned char *r,unsigned char *g,unsigned char *b,unsigned char *l,int n) {
    while(n--) {
	l[0] = r[0];
	l[1] = g[0];
	l[2] = b[0];
	l += 3; r++; g++; b++;
    }
}

typedef struct _ImageRec {
    unsigned short imagic;
    unsigned short type;
    unsigned short dim;
    unsigned short xsize, ysize, zsize;
    unsigned int min, max;
    unsigned int wasteBytes;
    char name[80];
    unsigned long colorMap;
    FILE *file;
    unsigned char *tmp; // , *tmpR, *tmpG, *tmpB;
    unsigned long rleEnd;
    unsigned int *rowStart;
    int *rowSize;
} ImageRec;

static void
ConvertShort(unsigned short *array, long length) {
    unsigned b1, b2;
    unsigned char *ptr;

    ptr = (unsigned char *)array;
    while (length--) {
	b1 = *ptr++;
	b2 = *ptr++;
	*array++ = (b1 << 8) | (b2);
    }
}

static void
ConvertLong(unsigned *array, long length) {
    unsigned long b1, b2, b3, b4;
    unsigned char *ptr;

    ptr = (unsigned char *)array;
    while (length--) {
	b1 = *ptr++;
	b2 = *ptr++;
	b3 = *ptr++;
	b4 = *ptr++;
	*array++ = (b1 << 24) | (b2 << 16) | (b3 << 8) | (b4);
    }
}

static void sgifread (void *buffer, int size, int count, FILE *stream)
{
  if (fread (buffer, size, count, stream) != (unsigned) count)
    mglError (MGL_ERROR_READING_IMAGE_DATA, "Error reading image data");
}


static ImageRec *ImageOpen(FILE *f)
{
    union {
	int testWord;
	char testByte[4];
    } endianTest;
    ImageRec *image;
    int swapFlag;
    int x;

    endianTest.testWord = 1;
    if (endianTest.testByte[0] == 1) {
	swapFlag = 1;
    } else {
	swapFlag = 0;
    }

    image = (ImageRec *)_malloc(sizeof(ImageRec));

    image->file = f;

    if (fread(image, 1, 12, image->file) != 12)
      imageError (MGL_ERROR_READING_SGI_HEADER, "Error reading SGI image header");

    if (swapFlag) {
    	ConvertShort(&image->imagic, 6);
    }

    image->tmp = (unsigned char *)_malloc(image->xsize*256);

    if ((image->type & 0xFF00) == 0x0100) {
	x = image->ysize * image->zsize * sizeof(unsigned);
	image->rowStart = (unsigned *)_malloc(x);
	image->rowSize = (int *)_malloc(x);
	image->rleEnd = 512 + (2 * x);
	fseek(image->file, 512, SEEK_SET);
	sgifread(image->rowStart, 1, x, image->file);
	sgifread(image->rowSize, 1, x, image->file);
	if (swapFlag) {
	    ConvertLong(image->rowStart, x/sizeof(unsigned));
	    ConvertLong((unsigned *)image->rowSize, x/sizeof(int));
	}
    }
    return image;
}

static void
ImageClose(ImageRec *image) {
    _free(image->tmp);
    _free(image);
}

static void
ImageGetRow(ImageRec *image, unsigned char *buf, int y, int z) {
    unsigned char *iPtr, *oPtr, pixel;
    int count;

    if ((image->type & 0xFF00) == 0x0100) {
	fseek(image->file, image->rowStart[y+z*image->ysize], SEEK_SET);
	sgifread(image->tmp, 1, (unsigned int)image->rowSize[y+z*image->ysize],
	      image->file);

	iPtr = image->tmp;
	oPtr = buf;
	while (1) {
	    pixel = *iPtr++;
	    count = (int)(pixel & 0x7F);
	    if (!count) {
		return;
	    }
	    if (pixel & 0x80) {
		while (count--) {
		    *oPtr++ = *iPtr++;
		}
	    } else {
		pixel = *iPtr++;
		while (count--) {
		    *oPtr++ = pixel;
		}
	    }
	}
    } else {
	fseek(image->file, 512+(y*image->xsize)+(z*image->xsize*image->ysize),
	      SEEK_SET);
	sgifread(buf, 1, image->xsize, image->file);
    }
}


unsigned char *
read_sgi(FILE *f, int *width, int *height, int *components, int alpha) {
    unsigned char *base, *lptr;
    unsigned char *rbuf, *gbuf, *bbuf, *abuf;
    ImageRec *image;
    int y;
    int d;

    image = ImageOpen(f);

  mglNote ("SGI image opened\n");

    if(!image)
	return NULL;
    (*width)=image->xsize;
    (*height)=image->ysize;
    d = (*components)=image->zsize;

    switch (d)
    {
      case 1:
        if (alpha == 1)
        {
          alphaExtend = 1;
          d = 2;
        }
        if (alpha >= 2)
        {
          if (mglVersionNumber < 11)
            d = 2;
        }
        break;
      case 3:
        if (alpha)
        {
          alphaExtend = 1;
          d = 4;
        }
        break;
      default:
        if (!alpha)
          d--;
        break;
    }

    base = (unsigned char *)_malloc(image->xsize*image->ysize*d*
				   sizeof(unsigned char));
    rbuf = (unsigned char *)_malloc(image->xsize*sizeof(unsigned char));
    gbuf = (unsigned char *)_malloc(image->xsize*sizeof(unsigned char));
    bbuf = (unsigned char *)_malloc(image->xsize*sizeof(unsigned char));
    abuf = (unsigned char *)_malloc(image->xsize*sizeof(unsigned char));
    memset (abuf, 255, image->xsize);
    if(!base || !rbuf || !gbuf || !bbuf)
      return NULL;
    lptr = base;
    for(y=0; y<image->ysize; y++) {
	if(image->zsize>=4) {
	    ImageGetRow(image,rbuf,y,0);
	    ImageGetRow(image,gbuf,y,1);
	    ImageGetRow(image,bbuf,y,2);
	    ImageGetRow(image,abuf,y,3);
      if (d == 4)
      {
	      rgbatorgba(rbuf,gbuf,bbuf,abuf,lptr,image->xsize);
      }
      else
      {
        rgbtorgb(rbuf,gbuf,bbuf,lptr,image->xsize);
      }
	    lptr += image->xsize*d;
	} else if(image->zsize==3) {
	    ImageGetRow(image,rbuf,y,0);
	    ImageGetRow(image,gbuf,y,1);
	    ImageGetRow(image,bbuf,y,2);
      if (d == 3)
	      rgbtorgb(rbuf,gbuf,bbuf,lptr,image->xsize);
      else
	      rgbatorgba(rbuf,gbuf,bbuf,abuf,lptr,image->xsize);
	    lptr += image->xsize*d;
	} else if(image->zsize==2) {
	    ImageGetRow(image,rbuf,y,0);
	    ImageGetRow(image,abuf,y,1);
      if (d == 2)
	      latola(rbuf,abuf,lptr,image->xsize);
      else
        bwtobw(rbuf,lptr,image->xsize);
	    lptr += image->xsize*d;
	} else {
	    ImageGetRow(image,rbuf,y,0);
      if (d == 1)
	      bwtobw(rbuf,lptr,image->xsize);
      else
	      latola(rbuf,abuf,lptr,image->xsize);
	    lptr += image->xsize*d;
	}
    }
    ImageClose(image);
    _free(rbuf);
    _free(gbuf);
    _free(bbuf);
    _free(abuf);

    *components = d;

    return (void *) base;
}



mglImage *loadSGI (FILE* f, int alpha)
{
  mglImage *image;
  int w;
  int h;
  int c;
  int type;
  unsigned char *data;

  data = read_sgi (f, &w, &h, &c, alpha);

  mglNote ("Creating image.\n");
  image = mglNewImage ();
  image->w = w;
  image->h = h;
  image->d = c;
//  image->c = c;
  image->data = data;

  switch (c)
  {
    case 1:
      type = GL_LUMINANCE;
      break;
    case 2:
      type = GL_LUMINANCE_ALPHA;
      break;
    case 3:
      type = GL_RGB;
      break;
    case 4:
      type = GL_RGBA;
      break;
    default:
      imageError (MGL_ERROR_UNSUPPORTED_SGI_IMAGE, "Detected unsupported SGI image file");
      return NULL;
  }
  image->type = type;
  image->yrev = 1;

  mglNote ("SGI image loaded successfully.\n");

  return image;
}

static void writef (void *buf, int size, FILE *f)
{
  if (fwrite (buf, size, 1, f) != 1)
    imageError (MGL_ERROR_WRITING_FILE, "Error writing to file. Possible cause - disk full");
}


/*---------------------------------------------------------------------------*/


static void saveTGA (FILE* f, mglImage *image)
{
//  TGA_ending ending = {0};
//  char ids [] = "TRUEVISION-XFILE";
  char bu [TGA_headerSize];
  char *b = bu;
  char* buf;
  char *data;
  TGA_header h;
  int size;

  memset (&h, 0, sizeof (TGA_header));

	h.image_width = (short)image->w;
	h.image_height = (short)image->h;
  if (image->yrev == 0)
  {
	  h.descriptor_bits = TGA_NORMALIMAGE; // y 0 je nahore
  }

  data  = (char *) image->data;
  buf = data;
  size = image->w * image->h;
//  h.descriptor_bits = 8;

  switch (image->type)
  {
    case GL_LUMINANCE_ALPHA:
      buf = _malloc (size);
      la2l (buf, data, size);
    case GL_LUMINANCE:
      h.pixel_bits = 8;
      h.image_type = 3;
      break;
    case GL_RGB:
      buf = _malloc (3*size);
      bgr2rgb (buf, data, size);
      h.pixel_bits = 24;
      h.image_type = 2;
      size *= 3;
      break;
//#ifdef GL_EXT_bgra
    case GL_BGR:
      h.pixel_bits = 24;
      h.image_type = 2;
//      h.palette_entry_size = 24;
      size *= 3;
      break;
//#endif
    case GL_RGBA:
      buf = _malloc (4*size);
      la2l (buf, data, size);
      size *= 4;
      h.pixel_bits = 32;
      //h.palette_entry_size = 32;
      h.image_type = 2;
      break;
//#ifdef GL_EXT_bgra
    case GL_BGRA:
      size *= 4;
      h.pixel_bits = 32;
      //h.palette_entry_size = 32;
      h.image_type = 2;
      break;
//#endif
    default:
      imageError (MGL_ERROR_UNSUPPORTED_TGA_IMAGE, "Unable to write TGA image - image file format not supported");
      break;
  }

  *b = h.id_length; b++;
  *b = h.palette_type; b++;
  *b = h.image_type; b++;
  *(unsigned short *)b = pcShort (h.first_color); b+=2;
  *(unsigned short *)b = pcShort (h.palette_colors); b+=2;
  *b = h.palette_entry_size; b++;
  *(unsigned short *)b = pcShort (h.left); b+=2;
  *(unsigned short *)b = pcShort (h.top); b+=2;
  *(unsigned short *)b = pcShort (h.image_width); b+=2;
  *(unsigned short *)b = pcShort (h.image_height); b+=2;
  *b = h.pixel_bits; b++;
  *b = h.descriptor_bits; b++;

  writef (bu, TGA_headerSize, f);
  writef (buf, size, f);
//  writef (&ending, sizeof (TGA_ending), f);
//  writef (ids, strlen (ids)+1, f);
  if (buf != data)
	  _free (buf);
}



#define WIDTHBYTES(bits)    (((bits) + 31) / 32 * 4)
#define BMP_HEADERSIZE (3 * 2 + 4 * 12)
#define BMP_BYTESPERLINE (width, bits) ((((width) * (bits) + 31) / 32) * 4)
#define BMP_PIXELSIZE(width, height, bits) (((((width) * (bits) + 31) / 32) * 4) * height)


static void saveBMP (FILE* f, mglImage *image)
{
	char bu [BMP_headerSize];
	char *b = bu;
	char* buf;
	char *data;
	BMP_header h;
	int size;

	// omezeni save BMP - neuklada 8-bitove obrazky
	// neuklada 32-bitove obrazky, pouze 24-bit, pokud je treba, zkonvertuje
	// neumi kompresi
	// predpoklada, ze obrazek je 4 point aligned (DWORD) aligned

	h.bi.size=40; 						// header size
	h.bi.width=image->w;
	h.bi.height=image->h;
	h.bi.planes=1;
	h.bi.bitCount=24;					// RGB encoded, 24 bit
	h.bi.compression=BI_RGB;			// no compression
	h.bi.sizeImage=0;
	h.bi.xPelsPerMeter=0;
	h.bi.yPelsPerMeter=0;
	h.bi.clrUsed=0;
	h.bi.clrImportant=0;

  data  = (char *) image->data;
  buf = data;
  size = image->w * image->h;
//  h.descriptor_bits = 8;

  switch (image->type)
  {
    case GL_RGB:
      buf = _malloc (3*size);
      bgr2rgb (buf, data, size);
      break;
//#ifdef GL_EXT_bgra
    case GL_BGR:
//      h.palette_entry_size = 24;
      break;
//#endif
    case GL_RGBA:
      buf = _malloc (3*size);
      bgra2rgb (buf, data, size);
      break;
//#ifdef GL_EXT_bgra
    case GL_BGRA:
		buf = _malloc (3*size);
		bgra2bgr (buf, data, size);
      break;
//#endif
    default:
      imageError (MGL_ERROR_UNSUPPORTED_BMP_IMAGE, "Unable to write BMP image - image file format not supported");
      break;
  }

  if (image->w % 4)
	  imageError (MGL_ERROR_UNSUPPORTED_BMP_IMAGE, "Unable to write BMP image - image width is not divisible by 4");

	size *= 3;

	h.bf.type = 0x4d42;
	h.bf.size = BMP_headerSize + WIDTHBYTES(image->w * 24) * image->h;	
	h.bf.reserved1 = h.bf.reserved2 = 0;
	h.bf.offBits = 54;

	*(unsigned short *)b = pcShort (h.bf.type); b+=2;
	*(unsigned long *)b = pcLong (h.bf.size); b+=4;
	*(unsigned short *)b = pcShort (h.bf.reserved1); b+=2;
	*(unsigned short *)b = pcShort (h.bf.reserved2); b+=2;
	*(unsigned long *)b = pcLong (h.bf.offBits); b+=4;
	*(unsigned long *)b = pcLong (h.bi.size); b+=4;
	*(unsigned long *)b = pcLong (h.bi.width); b+=4;
	*(unsigned long *)b = pcLong (h.bi.height); b+=4;
	*(unsigned short *)b = pcShort (h.bi.planes); b+=2;
	*(unsigned short *)b = pcShort (h.bi.bitCount); b+=2;
	*(unsigned long *)b = pcLong (h.bi.compression); b+=4;
	*(unsigned long *)b = pcLong (h.bi.sizeImage); b+=4;
	*(unsigned long *)b = pcLong (h.bi.xPelsPerMeter); b+=4;
	*(unsigned long *)b = pcLong (h.bi.yPelsPerMeter); b+=4;
	*(unsigned long *)b = pcLong (h.bi.clrUsed); b+=4;
	*(unsigned long *)b = pcLong (h.bi.clrImportant); b+=4;

  writef (bu, BMP_headerSize, f);
  writef (buf, size, f);

  if (buf != data)
	  _free (buf);
//  writef (&ending, sizeof (TGA_ending), f);
//  writef (ids, strlen (ids)+1, f);
}


/*---------------------------------------------------------------------------*/
/* Microsoft BMP */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Hlavni funkce */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/


void mglSaveImage (char *fileName, mglImage *image)
{
	FILE* f;
	char *s;

  curFileName = fileName;

	if ( ( f = fopen (fileName, "wb" ) ) == NULL )
  {
		imageError (MGL_ERROR_CREATING_FILE, "Cannot open file for writing");
		return;
	}

  s = strrchr (fileName, '.'); 
  if (s && !_stricmp (++s, "bmp"))
	  saveBMP (f, image);
  else
	// Zatim umim psat jen do targy
	saveTGA (f, image);

	fclose (f);
}


/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/



mglImage *mglLoadImage (char* fileName, int alpha)
{
  int type;
  FILE* f;
  mglImage *image;
  int i;
  int d;
  int size;
  unsigned char *p;

  if (alpha < 0)
    alpha = 1;

  alphaExtend = 0;

  curFileName = fileName;

  mglNote ("Opening ");
  mglNote (fileName);

  if ( ( f = fopen( fileName, "rb" ) ) == NULL )
  {
    imageError (MGL_ERROR_OPENING_FILE, "Error opening image file" );
    return NULL;
  }

  type = detectFormat (f);

  switch (type)
  {
    case MGL_IMAGEFORMAT_BMP:
      mglNote ("\nDetected image type is BMP\n");
      image = loadBMP (f, alpha);
      break;
    case MGL_IMAGEFORMAT_TGA:
      mglNote ("\nDetected image type is TGA\n");
      image = loadTGA (f, alpha);
      break;

    case MGL_IMAGEFORMAT_SGI:
      mglNote ("\nDetected image type is SGI\n");
      image = loadSGI (f, alpha);
      break;
    case MGL_IMAGEFORMAT_ERR:
    default:
      imageError (MGL_ERROR_UNSUPPORTED_IMAGE, "Unknown image format");
      return NULL;
  }

  fclose (f);

  if (alpha >= 2)
  {
    d = image->d;
    size = d * image->w * image->h;
    p = image->data;
    switch (d)
    {
      case 1:
        image->type = GL_INTENSITY;
        break;
      case 2:
        if (alphaExtend)
          for (i = 0; i < size; i += d)
          {
            p [i+1] = p [i];
          }
        break;
      case 4:
        if (alphaExtend)
        {
          for (i = 0; i < size; i += d)
          {
            p [i+3] = (p [i] + p [i+1] + p [i+2]) / 3;
          }
        }
    }
  }

  /*
  if (alpha == 2)
  {
    #ifdef GL_VERSION_1_1
      if (image->d == 1)
      {
        if (mglVersionNumber >= 11)
          image->type = GL_INTENSITY;
      }
    #endif

    if (image->d == 2)
    {
      p = (MGL_LAColor *) image->data;

      for (i = 0; i < image->w * image->h; i++)
      {
        p [i].a = p [i].l;
      }
    }
  }
  */

	return image;
}



typedef struct
{
  GLint x;
  GLint y;
  GLint w;
  GLint h;
} GL_viewport;

extern int mglRegionX;
extern int mglRegionY;


mglImage *mglGetRGBImage (int x, int y, int w, int h)
{
  GL_viewport	viewport;		/* Current viewport */
  mglImage *image;

  mglDoneState ();
 /*
  * Grab the current viewport...
  */
  glGetIntegerv (GL_VIEWPORT, (GLint *) &viewport);

  x += mglRegionX;
  y += mglRegionY;

  if (x < viewport.x)
    x = viewport.x;
  if (y < viewport.y)
    y = viewport.y;

  if (w > viewport.w)
    w = viewport.w - x;

  if (h > viewport.h)
    h = viewport.h - y;

  while ((w % 4) != 0)
    w--;

  if (w == 0)
    w = 4;

  image = mglCreateImage (w, h, 3);

  glFinish ();

  mglTestOpenGLError ();

  image->yrev = 1;

  image->type = mglExtBGRA ? GL_BGR : GL_RGB;

   // Read pixels from the framebuffer...
  glReadPixels (x, y, w, h, image->type, GL_UNSIGNED_BYTE,
               image->data);

  mglTestOpenGLError ();

  return image;
}
