
/*
 * MGL graphics library
 * Copyright (C) 1999-2000, ing. Marek Gayer

 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.

 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <mgl.h>
#include <math.h>

#ifndef GL_VERSION_1_1
#error OpenGL version 1.1 or above required to compile MGL library
#endif

#define MARKOVA_FREKVENCE 999999999

int mglLog = 0;

int mglTextureBufferSize;

// standardni virtualni souradnice
float mglVirtualWidth;
float mglVirtualHeight;
float mglVirtualStartX;
float mglVirtualStartY;
float mglVirtualEndX;
float mglVirtualEndY;

float mglCoordX (float x)
{
	return mglVirtualStartX + x * (mglVirtualEndX - mglVirtualStartX);
}

float mglUncoordX (float x)
{
	return (x - mglVirtualStartX) / (mglVirtualEndX - mglVirtualStartX);
}

float mglCoordY (float y)
{
	return mglVirtualStartY + y * (mglVirtualEndY - mglVirtualStartY);
}

float mglUncoordY (float y)
{
	return (y - mglVirtualStartY) / (mglVirtualEndY - mglVirtualStartY);
}


char *mglVendor;
char *mglRenderer;
char *mglVersion;
int mglVersionMajor;
int mglVersionMinor;
int mglVersionNumber;
char *mglExtensions;
//char *mglGLUVersion;
//char *mglGLUExtensions;

int mglAccelarated;
int mglMaxTextureSize;
int mglBitsPerPixel;
int mglZBufferBits;

//int mglExtABGR;
int mglExtBGRA;
int mglExtTexture;

static int infoOk;

extern void mglInitState (void);
extern void mglDoneState (void);



//-----------------------------------------------------------------------------

typedef struct
{
  GLint x;
  GLint y;
  GLint w;
  GLint h;
} GL_viewport;


static GLint swapbytes, lsbfirst, rowlength, skiprows, skippixels, alignment;
static GLint pack1, pack2, pack3, pack4;

extern void mglGetGLInfo ();
extern void mglInitInput (void);

extern void mglGetSizeRanges (void);


static void mglStorePixelAligment (void)
{
	/* mc_savecapture current pixel store state. */
	glGetIntegerv(GL_UNPACK_SWAP_BYTES, &swapbytes);
	glGetIntegerv(GL_UNPACK_LSB_FIRST, &lsbfirst);
	glGetIntegerv(GL_UNPACK_ROW_LENGTH, &rowlength);
	glGetIntegerv(GL_UNPACK_SKIP_ROWS, &skiprows);
	glGetIntegerv(GL_UNPACK_SKIP_PIXELS, &skippixels);
	glGetIntegerv(GL_UNPACK_ALIGNMENT, &alignment);

	glGetIntegerv(GL_PACK_ROW_LENGTH, &pack1);
	glGetIntegerv(GL_PACK_SKIP_ROWS, &pack2);
	glGetIntegerv(GL_PACK_SKIP_PIXELS, &pack3);
	glGetIntegerv(GL_PACK_ALIGNMENT, &pack4);

	/* Set desired pixel store state. */
	glPixelStorei(GL_UNPACK_SWAP_BYTES, GL_FALSE);
	glPixelStorei(GL_UNPACK_LSB_FIRST, GL_FALSE);
	glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
	glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
	glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);


	glPixelStorei (GL_PACK_ALIGNMENT, 4);	/* Force 4-byte alignment */
	glPixelStorei (GL_PACK_ROW_LENGTH, 0);
	glPixelStorei (GL_PACK_SKIP_ROWS, 0);
	glPixelStorei (GL_PACK_SKIP_PIXELS, 0);
}


static void mglSetupViewStub (void)
{
	if (!mglVirtualEndX && !mglVirtualStartX)
	{
		mglVirtualStartX = 0;
		mglVirtualEndX = mglVirtualWidth;
	}
	if (!mglVirtualEndY && !mglVirtualStartY)
	{
		mglVirtualStartY = mglVirtualHeight;
		mglVirtualEndY = 0;
	}

	mglDoneState ();

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	gluOrtho2D (mglVirtualStartX, mglVirtualEndX, mglVirtualStartY, mglVirtualEndY);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(0.0, 0.0, 0.0);
}

void mglSetupView (float left, float right, float bottom, float top)
{
	mglVirtualStartX = left; mglVirtualEndX = right; 
	mglVirtualStartY = bottom; mglVirtualEndY = top; 
	mglVirtualWidth = fabs (left - right); mglVirtualHeight = fabs (bottom - top);
	mglSetupViewStub ();
}

extern int mglWinWidth, mglWinHeight;

void mglSetupMGLMode (void)
{
  float size;
  GL_viewport	viewport;		/* Current viewport */

  mglDoneState ();

  if (!infoOk)
  {
    mglGetGLInfo ();
//    mglInitInput ();
  }

  glDisable (GL_DITHER);
	glDisable(GL_LIGHTING);
  glDisable(GL_FOG);
  glDisable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);
  glEnable (GL_CULL_FACE);
  glCullFace (GL_BACK);

  glDisable (GL_LOGIC_OP);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);

  glGetIntegerv (GL_VIEWPORT, (GLint *) &viewport);
  mglWindowWidth = viewport.w;
  mglWindowHeight = viewport.h;

  if (mglWinWidth && mglWinHeight)
  {
	glScissor (0, 0, mglWinWidth, mglWinHeight);
	glEnable (GL_SCISSOR_TEST);
  }

  //	glDisable (GL_SCISSOR_TEST);

  if (mglVirtualWidth == 0)
    mglVirtualWidth = mglWindowWidth;
  if (mglVirtualHeight == 0)
    mglVirtualHeight = mglWindowHeight;

  size = ((float)mglWindowWidth / mglVirtualWidth + (float)mglWindowHeight / mglVirtualHeight) / 2;
  if (size > 1)
  {
    mglSetPointSize (size);
    mglSetLineSize (size);
  }

  mglSetupViewStub ();

  //glDepthFunc(GL_LESS);

  // setups pixel aligment to byte ordering - standard mode, no word, double word aligment
  mglStorePixelAligment ();

  mglGetSizeRanges ();

  mglTestOpenGLError ();
}



extern int mglLastUsedTexture;


void mglBegin (void)
{
  mglDoneState ();

  glPushAttrib(GL_ENABLE_BIT | GL_VIEWPORT_BIT | GL_TRANSFORM_BIT);

  glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glMatrixMode(GL_MODELVIEW);
  glPushMatrix();

  mglLastUsedTexture = -1;

  mglSetupMGLMode ();

  mglInitState ();

  mglTestOpenGLError ();
}

static void mglRestorePixelAligment (void)
{
	/* Restore current pixel store state. */
	glPixelStorei(GL_UNPACK_SWAP_BYTES, swapbytes);
	glPixelStorei(GL_UNPACK_LSB_FIRST, lsbfirst);
	glPixelStorei(GL_UNPACK_ROW_LENGTH, rowlength);
	glPixelStorei(GL_UNPACK_SKIP_ROWS, skiprows);
	glPixelStorei(GL_UNPACK_SKIP_PIXELS, skippixels);
	glPixelStorei(GL_UNPACK_ALIGNMENT, alignment);

	glPixelStorei(GL_UNPACK_ROW_LENGTH, pack1);
	glPixelStorei(GL_UNPACK_SKIP_ROWS, pack2);
	glPixelStorei(GL_UNPACK_SKIP_PIXELS, pack3);
	glPixelStorei(GL_UNPACK_ALIGNMENT, pack4);
}


void mglEnd (void)
{
  mglDoneState ();

  glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glPopAttrib();

  mglTestOpenGLError ();
}



static int mglCheckExtension (char *extName)
{
  /*
  ** Search for extName in the extensions string.  Use of strstr()
  ** is not sufficient because extension names can be prefixes of
  ** other extension names.  Could use strtok() but the constant
  ** string returned by glGetString can be in read-only memory.
  */

  char *p = (char *) mglExtensions;
  char *end;
  int extNameLen;

  extNameLen = strlen(extName);
  end = p + strlen(p);

  while (p < end)
  {
    int n = strcspn(p, " ");
    if ((extNameLen == n) && (strncmp(extName, p, n) == 0))
	{
      return 1;
	}
	p += (n + 1);
  }
  return 0;
}


static int mglGetAcceleration (void)
{
  return 0;
}


static void mglPrintGLInfo (void)
{
  char s [64000];

  sprintf (s, "GL_VENDOR: %s\n"
	          "GL_RENDERER: %s\n"
		      "GL_VERSION: %s\n"
          "GL_EXTENSIONS: %s\n"
          "Screen resolution is %i x %i, %i bpp\n"
          "Maximum texture size is %i x %i pixels.\n"
          "Detected API version: %i.%i\n"
          "Texture objects: %s\n"
          "BGRA extension: %s\n",
  mglVendor, mglRenderer, mglVersion, mglExtensions,
  mglScreenWidth, mglScreenHeight, mglBitsPerPixel,
  mglMaxTextureSize, mglMaxTextureSize, mglVersionMajor, mglVersionMinor,
  mglExtTexture ? "Yes" : "No", mglExtBGRA ? "Yes" : "No");

  mglNote (s);
}

static int mglVersionCheck (void)
{
  char s [333];

  if (sscanf (mglVersion, "%i.%i", &mglVersionMajor, &mglVersionMinor) != 2)
  {
    sprintf (s, "Unknown OpenGL version detected (%s).", mglVersion);
    mglError (MGL_ERROR_UNKNOWN_OPENGL_VERSION, s);
  }

#ifndef GL_VERSION_1_2
  mglVersionMinor = 1;
#endif

#ifndef GL_VERSION_1_1
  mglVersionMinor = 0;
#endif

//  mglVersionMinor = 0; // for debug ! remove ...

  return 10 * mglVersionMajor + mglVersionMinor;
}



/*
int mglGetTextureBufferSize (void)
{
  mglImage *image;
  mglTexture *texture [256];
  int i;
  int result;

  image = mglCreateImage (512, 512, 4);
  image->type = GL_RGBA;
  for (i = 0; i < 32; i++)
  {
    texture [i] = mglMakeTexture (image, 0);
    if (!texture [i]->resident)
    {
      mglDeleteTexture (texture [i]);
      break;
    }
  }
  result = mglResidentTexturesMemory;

  for (i-- ; i >= 0 ; i--)
    mglDeleteTexture (texture [i]);

  return result;
}
*/

static int mglProxySupported (int w, int h)
{
  int proxyComponents;

  glTexImage2D(GL_PROXY_TEXTURE_2D, 0, GL_RGBA8,
                w, h, 0,
                GL_RGBA, GL_UNSIGNED_BYTE, NULL);
  glGetTexLevelParameteriv(GL_PROXY_TEXTURE_2D, 0,
                            GL_TEXTURE_INTERNAL_FORMAT, &proxyComponents);

  if (proxyComponents == GL_RGBA8)
     return 1;
  else
    return 0;

  mglTestOpenGLError ();
}

static void mglGetMaxTextureSize (void)
{
  /*
  int i;

  mglNote ("Checking proxy ");

  if (mglExtTexture)
  {
    i = 16;

    while (mglProxySupported (i, i))
    {
      mglNote (".");
      i += i;
    }

    mglMaxTextureSize = i / 2;
  }
  else
  */
  {
    glGetIntegerv (GL_MAX_TEXTURE_SIZE, &mglMaxTextureSize);
  }
}


void mglGetGLInfo (void)
{
  int redBits;
  int greenBits;
  int blueBits;
  int alphaBits;
  int size;

  mglDoneState ();

  // checkup whether 32-bit compiler used
  size = sizeof (char);
  if (size != 1)
    mglError (MGL_ERROR_COMPILER_OPTIONS, "Invalid compiler settings detected (sizeof (char) must be 1).");
  size = sizeof (short);
  if (size != 2)
    mglError (MGL_ERROR_COMPILER_OPTIONS, "Invalid compiler settings detected (sizeof (short) must be 2).");
  size = sizeof (int);
  if (size != 4)
    mglError (MGL_ERROR_COMPILER_OPTIONS, "Invalid compiler settings detected (sizeof (int) must be 4).");
  if (sizeof (float) != sizeof (GLfloat))
    mglError (MGL_ERROR_COMPILER_OPTIONS, "Invalid compiler settings detected (sizeof (float) must be sizeof (GLfloat).");

  infoOk = 1;

  mglNote ("Getting OpenGL and GLU strings\n");
  mglVendor = _strdup ((char *)glGetString (GL_VENDOR));
  mglRenderer = _strdup ((char *) glGetString (GL_RENDERER));
  mglVersion = _strdup ((char *) glGetString (GL_VERSION));
  mglExtensions = _strdup ((char *) glGetString (GL_EXTENSIONS));

  if ((mglVendor == NULL) || (mglRenderer == NULL) || (mglVersion == NULL)
       || mglExtensions == NULL)
    mglError (MGL_ERROR_NO_OPENGL_DRIVERS, "No valid OpenGL drivers detected.");

  mglVersionNumber = mglVersionCheck ();
//  if (mglVersionCheck () < 11)
//  {
    //mglError ("MGL graphics library requires OpenGL version 1.1 or above.\n");
//  }

//  mglGLUVersion = ""; // (char *) gluGetString (GLU_VERSION);
//  mglGLUExtensions = (char *) gluGetString (GLU_EXTENSIONS);
//  if ((mglGLUVersion == NULL) || (mglGLUExtensions == NULL))
    //mglError (MGL_ERROR_NO_GLU_DRIVERS, "No valid GLU drivers detected.");

  mglAccelarated = mglGetAcceleration ();

	glGetIntegerv( GL_DEPTH_BITS, &mglZBufferBits);
	// glGetIntegerv( GL_STENCIL_BITS, &stencilBits );

	glGetIntegerv( GL_RED_BITS, &redBits );
	glGetIntegerv( GL_GREEN_BITS, &greenBits );
	glGetIntegerv( GL_BLUE_BITS, &blueBits );
	glGetIntegerv( GL_ALPHA_BITS, &alphaBits );
  mglBitsPerPixel = redBits + greenBits + blueBits + alphaBits;

  // checkup for the standard Open GL extensions used by MGL library
//  mglExtABGR = 0;
  mglExtBGRA = 0;
  mglExtTexture = 0;

  mglNote ("Checking extensions\n");

#ifdef GL_EXT_bgra
  mglExtBGRA = mglCheckExtension ("GL_EXT_bgra");
#endif

  if (mglVersionNumber >= 12)
  {
    mglExtBGRA = 1;
  }

/*
#ifdef GL_EXT_abgr
  mglExtABGR = mglCheckExtension ("GL_EXT_abgr");
#endif
*/

  if (mglVersionNumber < 11)
  {
    mglExtTexture = mglCheckExtension ("GL_EXT_texture_object");
  }
  else
    mglExtTexture = 1;

//  mglTextureBufferSize = mglGetTextureBufferSize ();

  mglNote ("Getting maximum texture size\n");

  mglGetMaxTextureSize ();

  mglNote ("Testing OpenGL error state\n");
  mglTestOpenGLError ();

//  mglExtTexture = 1;

  mglPrintGLInfo ();
}

/*
void mglSetVirtualWindowSize (int width, int height)
{
  mglVirtualWidth=width;
  mglVirtualHeight=height;
}
*/
