
/*
 * MGL graphics library
 * Copyright (C) 1999-2000, ing. Marek Gayer

 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.

 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <mgl.h>

#if !defined(GL_VERSION_1_1)
#if defined(GL_EXT_texture_object)
#define glBindTexture(A,b)     glBindTextureEXT(A,b)
#define glGenTextures(A,b)     glGenTexturesEXT(A,b)
#define glDeleteTextures(A,b)  glDeleteTexturesEXT(A,b)
#else
#define glBindTexture(A,b)
#define glGenTextures(A,b)
#define glDeleteTextures(A,b)
#define glPrioritizeTextures(A,b,C)
#endif
#endif

extern int mglExtTexture;


float mglTextureZoomX=1;
float mglTextureZoomY=1;

int mglResidentTexturesTotal;
int mglResidentTexturesMemory;
int mglLocalTexturesTotal;
int mglLocalTexturesMemory;

int mglLastUsedTexture;

mglColor mglTextureColor [4];

void mglSetTextureColors (mglColor color)
{
  int i;

  for (i = 0; i < 4; i++)
  {
    mglTextureColor [i] = color;
  }
}


void mglSetTextureZoom (float zoom)
{
  mglTextureZoomX = zoom;
  mglTextureZoomY = zoom;
}


void mglDeleteTexture (mglTexture *texture)
{
  GLuint tex;
  int size;

  mglDoneState ();

  if (!texture)
    return;

  size = texture->image.w*texture->image.h*texture->image.d;

  if (texture->residancechecked == 2)
  {
    mglResidentTexturesTotal--;
    mglResidentTexturesMemory-=size;
  }
  
  if (texture->residancechecked == 1)
  {
	  mglLocalTexturesTotal--;
	  mglLocalTexturesMemory-=size;
  }

  tex = texture->name;

  if (mglExtTexture)
    glDeleteTextures (1, &tex);
  else
    glDeleteLists (tex, 1);

  texture->name = 0;

  /*
  unsigned char **p;

  if (texture->image.data)
  {
    _free (texture->image.data);
    p = texture->image.mipmap;
    if (p)
    {
      while (*p)
      {
        _free (*p);
        p++;
      }
    }
  }
  */

  _free (texture);
}

static int checkSize (int x)
{
  int y;

  y = 1;

  while (y <= x)
  {
    if (x == y)
      return 1;

    y = 2 * y;
    if (y > mglMaxTextureSize)
      return 0;
  }

  return 0;
}


mglTexture *mglMakeTexture (mglImage *image, int mode)
{
  mglTexture *texture;
  GLclampf priority = 999;
  GLuint texName;
  int list;
  char s [100];
  int min, mag;
  int mipmaps;
  int level;
  int w;
  int h;
  int c;
  int t;
  unsigned char *mipmap;

  min = mode;

  mglDoneState ();

  switch (mode)
  {
    case GL_LINEAR:
      mipmaps = 0;
      mag = GL_LINEAR;
      break;
    case GL_NEAREST:
      mipmaps = 0;
      mag = GL_NEAREST;
      break;
    case GL_NEAREST_MIPMAP_NEAREST:
      mipmaps = 1;
      mag = GL_NEAREST;
      break;
    case GL_LINEAR_MIPMAP_LINEAR:
      mipmaps = 1;
      mag = GL_LINEAR;
      break;
    case GL_LINEAR_MIPMAP_NEAREST:
      mipmaps = 1;
      mag = GL_LINEAR;
      break;
    case GL_NEAREST_MIPMAP_LINEAR:
      mipmaps = 1;
      mag = GL_NEAREST;
      break;
    default:
      sprintf (s, "Invalid texture mode (%i) specified.", mode);
      mglError (MGL_ERROR_UNKNOWN_TEXTURE_MODE, s);
      return NULL;
  }

  texture = (mglTexture *) _malloc (sizeof (mglTexture));

  if ((image->h > mglMaxTextureSize) || (image->w > mglMaxTextureSize))
  {
    sprintf (s, "Texture size is too big. Requested %i x %i, supported %i x %i.", image->w, image->h, mglMaxTextureSize, mglMaxTextureSize);
    mglError (MGL_ERROR_TOO_BIG_TEXTURE, s);
  }

  if ((!checkSize (image->h)) || (!checkSize (image->w)))
  {
    sprintf (s, "Invalid width or height of texture (%i x %i). Both must be power of 2.", image->h, image->w);
    mglError (MGL_ERROR_TEXTURE_SIZE, s);
  }

  memcpy (&texture->image, image, sizeof (mglImage));
  texture->image.data = NULL;
  texture->image.mipmap = NULL;
  texture->residancechecked = 0;

  if (mglExtTexture)
  {
    glGenTextures(1, &texName);
    texture->name = texName;
    glPrioritizeTextures (1, &texName, &priority);
//    mglTestOpenGLError ();
    glBindTexture (GL_TEXTURE_2D, texName);
  }
  else
  {
    list = glGenLists (1);
    texture->name = list;
//    mglTestOpenGLError ();
    glNewList (list, GL_COMPILE);
  }
//  texture->priority = mglTexturePriority;

  c = image->d;
  t = image->type;
  if (t == GL_INTENSITY)
  {
    t = GL_LUMINANCE;
    c = GL_INTENSITY8;
  }

  // setups 2d texture
	// with mipmap level 0
	// the texture contains 3 color components
	// width and height of 2 pow n
  // width of border is 0
	// format of the texture is RGB (24 bytes) and color values of 0 - 255
  // texture is stored in stored in the client memory

  mglNote ("Making texture\n");

  if (!mipmaps)
    glTexImage2D (GL_TEXTURE_2D, 0, c, image->w, image->h,
				          0, t, GL_UNSIGNED_BYTE, image->data);
  else
  {
    if (image->mipmap == NULL)
    {
      gluBuild2DMipmaps(GL_TEXTURE_2D, c, image->w, image->h,
                      t, GL_UNSIGNED_BYTE, image->data);
    }
    else
    {
      level = 0;
      w = 2 * image->w;
      h = 2 * image->h;
      mipmap = image->data;

      while ((w > 1) || (h > 1))
      {
        if (w > 1) w /= 2;
        if (h > 1) h /= 2;
        if (mipmap == NULL)
        {
          sprintf (s, "Missing mipmap of level %i (%i x %i size)", level, w ,h);
          mglError (MGL_ERROR_MISSING_MIPMAP, s);
        }
        glTexImage2D (GL_TEXTURE_2D, level, c, w, h,
				  0, t, GL_UNSIGNED_BYTE, mipmap);

        mipmap = image->mipmap [level];
        level++;
      }
    }
  }

  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT); // or GL_CLAMP
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, mag); // GL_LINEAR
  if (!mipmaps)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, min); // GL_LINEAR
  else
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, min); // GL_NEAREST

  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE); // GL_BLEND GL_REPLAce
  // glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL); // GL_BLEND GL_REPLAce

  if (!mglExtTexture)
  {
    glEndList();
  }

  mglTestOpenGLError ();

  mglLastUsedTexture = -1;

  return texture;
}

mglTexture *mglLoadTexture (char *fileName, int imageMode, int textureMode)
{
  mglImage *image;
  mglTexture *texture;

  image = mglLoadImage (fileName, imageMode);
  texture = mglMakeTexture (image, textureMode);
  mglDeleteImage (image);
  texture->image.data = NULL;

  return texture;
}


void mglChangeTexture (mglTexture *texture)
{
  if (mglLastUsedTexture != texture->name)
  {
    mglLastUsedTexture = texture->name;

    mglDoneState ();

    if (mglExtTexture)
      glBindTexture (GL_TEXTURE_2D, texture->name);
    else
      glCallList (texture->name);

    mglTestOpenGLError ();
  }
}

extern void mglSetColor (mglColor color);


void mglDrawTexturedRectangle (mglTexture *texture, float x, float y,
                                float sizeX, float sizeY)
{
  mglDrawTexturedEngine (0, 0, 1.0 / (float)mglTextureZoomX, 1.0 / (float)mglTextureZoomY,
    texture, x, y, x + sizeX, y + sizeY);
}

void mglDrawTexture (mglTexture *texture, float x, float y)
{
  float maxx;
  float maxy;

  maxx = x + mglTextureZoomX * texture->image.w;
  maxy = y + mglTextureZoomY * texture->image.h;

  mglDrawTexturedEngine (0, 0, 1, 1, texture, x, y, maxx, maxy);
}



static float getTextureCord (int texture_size, int cord, int reversed)
{
  double result;
  char s [100];

  if (cord > texture_size)
  {
    sprintf (s, "Bad subtexture coordinates specified (%i of maximum texture length %i).", cord, reversed);
    mglError (MGL_ERROR_SUBTEXTURE_COORDINATES, s);
  }

  result = ((double)cord) / texture_size;
  if (reversed)
    return 1.0f - result;
  else
    return result;
}


void mglDrawSubTexture (mglTexture *texture, float x, float y, int fromX, int fromY,
                         int sizeX, int sizeY)
{
  float maxx;
  float maxy;
  float tx1, tx2, ty1, ty2;
  int rx, ry;

  rx = texture->image.xrev;
  ry = texture->image.yrev;

  maxx = x + mglTextureZoomX * sizeX;
  maxy = y + mglTextureZoomY * sizeY;

  tx1 = getTextureCord (texture->image.w, rx*sizeX + fromX, rx);
  tx2 = getTextureCord (texture->image.w, !rx*sizeX + fromX, rx);
  ty1 = getTextureCord (texture->image.h, ry*sizeY + fromY, ry);
  ty2 = getTextureCord (texture->image.h, !ry*sizeY + fromY, ry);

  mglDrawTexturedEngine (tx1, ty1, tx2, ty2, texture, x, y, maxx, maxy);
}


