//***************************************************************************
// MGLDraw 1.0
// (c) 2000 Emil Aubrecht
//***************************************************************************

//---------------------------------------------------------------------------
#include "kresleni.h"
//---------------------------------------------------------------------------

TElementBod::TElementBod(int ax, int ay, mglColor abarva) {

  x = ax;
  y = ay;
  barva = abarva;
  sirka = 10;
};
//---------------------------------------------------------------------------

void TElementBod::Zobraz() {

  mglPointColor = barva;
  mglSetPointSize(sirka > 0 ? sirka : 1.0);
  mglDrawPoint(x, y);
};
//---------------------------------------------------------------------------

void TElementBod::Save(FILE *soubor){

  if (soubor) {
    fprintf(soubor, "\n");

    fprintf(soubor, "  mglPointColor.r = %g;", barva.r);
    fprintf(soubor, " mglPointColor.g = %g;", barva.g);
    fprintf(soubor, " mglPointColor.b = %g;", barva.b);
    fprintf(soubor, " mglPointColor.a = %g;\n", barva.a);

    fprintf(soubor, "  mglSetPointSize(%g);\n", sirka);

    fprintf(soubor, "  mglDrawPoint (x + %i, y + %i);\n", VX(x), VY(y));
  };
};
//---------------------------------------------------------------------------

TElementCara::TElementCara(int ax[3], int ay[3], mglColor abarva[3]) {
  int i;

  for (i = 0; i<2; i++) {
    x[i] = ax[i];
    y[i] = ay[i];
    barva[i] = abarva[i];
  };
  sirka = 2;
};
//---------------------------------------------------------------------------

void TElementCara::Zobraz() {

  mglLineColor[0] = barva[0];
  mglLineColor[1] = barva[1];
  mglSetLineSize(sirka);
  mglDrawLine (x[0], y[0], x[1], y[1]);
};
//---------------------------------------------------------------------------

void TElementCara::Save(FILE *soubor){
  int i;

  if (soubor) {
    fprintf(soubor, "\n");
    for (i=0; i<2; i++) {
      fprintf(soubor, "  mglLineColor[%i].r = %g;", i, barva[i].r);
      fprintf(soubor, " mglLineColor[%i].g = %g;", i, barva[i].g);
      fprintf(soubor, " mglLineColor[%i].b = %g;", i, barva[i].b);
      fprintf(soubor, " mglLineColor[%i].a = %g;\n", i, barva[i].a);
    };
    fprintf(soubor, "  mglSetLineSize(%g);\n", sirka);

    fprintf(soubor, "  mglDrawLine(x + %i, y + %i, x + %i, y + %i);\n", VX(x[0]), VY(y[0]), VX(x[1]), VY(y[1]));
  };
};

//---------------------------------------------------------------------------

TElementTrojuhelnik::TElementTrojuhelnik(int ax[3], int ay[3], mglColor abarva[3]) {
  int i;

  for (i = 0; i<3; i++) {
    x[i] = ax[i];
    y[i] = ay[i];
    barva[i] = abarva[i];
  };
  vyplneny = false;
};
//---------------------------------------------------------------------------

void TElementTrojuhelnik::Zobraz() {

  mglTriangleColor[0] = barva[0];
  mglTriangleColor[1] = barva[1];
  mglTriangleColor[2] = barva[2];

  mglSetLineSize(1.0);

  if (vyplneny)
    mglDrawFilledTriangle(x[0], y[0], x[1], y[1], x[2], y[2]);
  else
    mglDrawTriangle(x[0], y[0], x[1], y[1], x[2], y[2]);
};
//---------------------------------------------------------------------------

void TElementTrojuhelnik::Save(FILE *soubor){
  int i;

  if (soubor) {
    fprintf(soubor, "\n");

    for (i=0; i<3; i++) {
      fprintf(soubor, "  mglTriangleColor[%i].r = %g;", i, barva[i].r);
      fprintf(soubor, " mglTriangleColor[%i].g = %g;", i, barva[i].g);
      fprintf(soubor, " mglTriangleColor[%i].b = %g;", i, barva[i].b);
      fprintf(soubor, " mglTriangleColor[%i].a = %g;\n", i, barva[i].a);
    };

    fprintf(soubor, "  mglSetLineSize(1);\n");

    fprintf(soubor, "  mglDraw");
    if (vyplneny)
      fprintf(soubor, "Filled");
    fprintf(soubor, "Triangle(x + %i, y + %i, x + %i, y + %i, x + %i, y + %i);\n", VX(x[0]), VY(y[0]), VX(x[1]),
            VY(y[1]), VX(x[2]), VY(y[2]));
  };
};
//---------------------------------------------------------------------------

TElementObdelnik::TElementObdelnik(int ax[3], int ay[3], mglColor abarva[3]) {
  int i;

  for (i = 0; i<2; i++) {
    x[i] = ax[i];
    y[i] = ay[i];
    barva[i] = abarva[i];
  };
  vyplneny = false;
};
//---------------------------------------------------------------------------

void TElementObdelnik::Zobraz() {

  mglRectangleColor[0] = barva[0];
  mglRectangleColor[1] = barva[0];
  mglRectangleColor[2] = barva[1]; //****
  mglRectangleColor[3] = barva[1];

  mglSetLineSize(1.0);

  if (vyplneny)
    mglDrawFilledRectangle(x[0], y[0], x[1], y[1]);
  else
    mglDrawRectangle(x[0], y[0], x[1], y[1]);
};
//---------------------------------------------------------------------------

void TElementObdelnik::Save(FILE *soubor){
  int i;

  if (soubor) {
    fprintf(soubor, "\n");

    for (i=0; i<4; i++) {
      fprintf(soubor, "  mglRectangleColor[%i].r = %g;", i, barva[i].r);
      fprintf(soubor, " mglRectangleColor[%i].g = %g;", i, barva[i].g);
      fprintf(soubor, " mglRectangleColor[%i].b = %g;", i, barva[i].b);
      fprintf(soubor, " mglRectangleColor[%i].a = %g;\n", i, barva[i].a);
    };

    fprintf(soubor, "  mglSetLineSize(1);\n");

    fprintf(soubor, "  mglDraw");
    if (vyplneny)
      fprintf(soubor, "Filled");
    fprintf(soubor, "Rectangle(x + %i, y + %i, %i, %i);\n", VX(x[0]), VY(y[0]), VX(x[1]), VY(y[1]));
  };
};
//---------------------------------------------------------------------------

